/***************************************************************************
 *                                                                         *
 *   basem.h     (begin: Mar 05 2003)                                      *
 *                                                                         *
 *   Parallel IQPNNI - Important Quartet Puzzle with NNI                   *
 *                                                                         *
 *   Copyright (C) 2003-2004 by Le Sy Vinh, Arndt von Haeseler             *
 *   {vinh,minh}@cs.uni-duesseldorf.de                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef BASEM_H
#define BASEM_H

#include "constant.h"
#include "mat.h"
#include "vec.h"
#include "dvec20.h"
#include "dmat20.h"
#include "brent.h"

const double SCALE_RATE = 0.01;
const double SCALE_RATE_CODON = 0.1;

/**
	this is the state model for all current models, all current models inherit from stateM
*/
class BaseM : public Brent {
public:
	/**
		constructor function
	*/
	BaseM();

	/**
		set the model type
	*/
	void setModel (MODEL modelType);

	/**
		return TRUE if model is a DNA model 
	*/
	int isDNA ();

	/**
		set the transition transvertion ratio
	*/
	void setTsTvRatio (const double tsTvRatio);

	/**
		set the pyridimine/purine ratio
	*/
	void setPyPuRatio (const double pyPuRatio);
	
		
	/**
		set the state frequencies of this alignment
	*/
	void setStateFrq (DVec20 &stateFrqArr);

	/**
		set parameters for DNA Model
	*/
	void setGenPam (const double tsAG, const double tsCT,
	                const double tvAC, const double tvAT, const double tvCG, const double tvGT);

	/**
		get parameters for DNA Model
	*/
	void getGenPam (double &tsAG,  double &tsCT,
	                double &tvAC,  double &tvAT, double &tvCG, double &tvGT);

	/**
		get the state frequencies of this alignment
	*/
	void getStateFrq (DVec20 &stateFrqArr);

	/**
		get the Transition/Transversion Ratio
	*/
	virtual double getTsTvRatio ();
	
	/**
		get the pyridimine/purine ratio
	*/
	virtual double getPyPuRatio ();

	/**
		compute the probability of changing from stateNo1 into stateNo2 after a period time of time
	*/
	virtual double cmpProbChange (const int stateNo1, const int stateNo2, const double time);
	
	/**
		compute the probability changes matrix
	*/
	virtual void cmpProbChange (const double time, DMat20 &probMat);
	
	/**
		compute the probability change matrix, for codon-based Nielsen Yang 98 Model
	*/
	virtual void cmpProbChange (const double time, const int classNo, DMat20 &probMat);
	
	/** 
		compute the probability changes vector
	*/
	virtual void cmpProbChange (const int stateNo, const double time, DVec20 &probVec);


	/**
		compute the 1st and 2nd derivatives of log likelihood
	*/
    virtual void cmpProbChangeDerivatives (const int stateNo, const double brLen, DVec20 &probVec, DVec20 &derv1, DVec20 &derv2);
	
	/**
		compute the 1st and 2nd derivatives of log likelihood
	*/
    virtual void cmpProbChangeDerivatives (const double brLen, DMat20 &probMat, DMat20 &derv1, DMat20 &derv2);
	
	
	/**
		compute the 1st and 2nd derivatives of log likelihood, for codon-based Nielsen Yang 98 Model
	*/
    virtual void cmpProbChangeDerivatives (const double brLen, const int classNo, DMat20 &probMat, DMat20 &derv1, DMat20 &derv2);
	
	/**
		all things are inited here
	*/
	virtual void init ();

	/**
		optimize parameters
	*/
	virtual bool optPam (const PAM_TYPE tsTvRatioType, const PAM_TYPE pyPuRatioType);
	
	/**
		optimize parameters
		@return TRUE of some parameters changed
	*/
	virtual bool optPam ();

	
	void elmhes(DMat20 a, Vec<int> &ordr, int n);
	void eltran(DMat20 a, DMat20 zz, Vec<int> &ordr, int n);
	void mcdiv(double ar, double ai, double br, double bi,
	           double *cr, double *ci);

	void hqr2(int n, int low, int hgh, DMat20 h,
	          DMat20 zz, DVec20 &wr, DVec20 &wi);
			  
	void onepamratematrix(DMat20 a, DVec20 stateFrq, int num_state);
	void eigensystem(DVec20 &eval, DMat20 &evec, int num_state);
	void luinverse(DMat20 inmat, DMat20 imtrx, int size);
	void checkevector(DMat20 evec, DMat20 ivec, int nn);
	
	
	/**
		get the rate parameters and frequency vector
	*/	
	virtual void rtfdata (DMat20 &a, DVec20 &forg, int num_state);

	void writeMat (DMat20 &rate, int num_state);
	void writeInt (Vec<int> &vec, int num_state);
	void writeDouble (DVec20 &vec, int num_state);
	
	void calcEigenCoeff(int num_state);

	/**
	*/	
	void tranprobmat(int num_state);
	
	/**
		release all memmory for this class
	*/
	void release ();

	/**
		the destructor function
	*/
	virtual ~BaseM();


	/************************************************************************************
	************************************************************************************/
	/**
		the total number of substitutions per unit time
	*/
	double total_num_subst;
	
protected :

	/**
		model type
	*/
	MODEL modelType_;

	/**
		number of states, 4 for DNA, 20 for Amino acid and 61 for Codon
	*/
	int nState_;
	
	/**
		active parameter, used in optimization phase
	*/
	MODEL_PAM_TYPE isActPam_;
	
	/**
		the transition transvertion ratio
	*/
	double tsTvRatio_;
	
	/**
		the pyridimine/purine ratio
	*/
	double pyPuRatio_;

	/**
		A -> G ratio
	*/
	double tsAG_;
	
	/**
		C -> T ratio
	*/
	double tsCT_;

	/**
		A -> C ratio
	*/
	double tvAC_;
	
	/**
		A -> T ratio
	*/
	double tvAT_;
	
	/**
		C -> G ratio
	*/
	double tvCG_;
	
	/**
		G -> T ratio
	*/
	double tvGT_;

	/**
		the state frequency list
	*/
	DVec20 stateFrqArr_;
	
	//double tsTvRatioError_;

	/**
		unit rate matrix
	*/
	DMat20 unitRateMat_;
	
	/**
		eigenvectors matrix
	*/
	DMat20 eigVecMat_;
	
	/**
		inversion of eigenvectors matrix
	*/
	DMat20 invEigVecMat_;
	
	/**
		eigenvalues 
	*/
	DVec20 eigValArr_;
	
	/**
		eigen coefficient, for 1st derivative, for 2nd derivative
	*/
	DMat20 eigen_coeff[MAX_STATE_NUM], eigen_coeff_derv1[MAX_STATE_NUM], eigen_coeff_derv2[MAX_STATE_NUM];
	
	DCube20 eigen_coeff_vec[MAX_NUM_CLASS], eigen_coeff_derv1_vec[MAX_NUM_CLASS], eigen_coeff_derv2_vec[MAX_NUM_CLASS];
	DVec20 eigValArr_vec[MAX_NUM_CLASS];
	
	/**
		compute the substitution rate per unit time for this model
	*/
	double cmpSubRate ();

	/**
		cmp the state frequencies for this alignment
	*/
	virtual void cmpStateFrq ();

}
; //end of stateM class

#endif
